/* -*- Mode: C -*- */


#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "libical/icalstringset.h"
#include "icalstringsetimpl.h"
#include "icalparser.h"
#include <errno.h>
#ifndef WIN32
//#include <unistd.h> /* for stat, getpid */
#else
#include <io.h>
#include <share.h>
#endif
#include <stdlib.h>
#include <string.h>

#ifdef WIN32
#include <sys/stat.h> /* for stat */
#include <fcntl.h> /* for fcntl */
#define snprintf	_snprintf
#define strcasecmp	stricmp

#define _S_ISTYPE(mode, mask)  (((mode) & _S_IFMT) == (mask))

#define S_ISDIR(mode)    _S_ISTYPE((mode), _S_IFDIR)
#define S_ISREG(mode)    _S_ISTYPE((mode), _S_IFREG)
#endif

/** Default options used when NULL is passed to icalset_new() **/
icalstringset_options icalstringset_options_default = { 0 };

icalset* icalstringset_new(const char* icalendar_str)
{
	return icalset_new(ICAL_STRING_SET, icalendar_str, &icalstringset_options_default);
}

icalset* icalstringset_init(icalset *set, const char* icalendar_str, void* options_in)
{
	icalstringset_options *options = (options_in) ? options_in : &icalstringset_options_default;
	icalstringset *fset = (icalstringset*) set;
	icalerror_clear_errno();
	icalerror_check_arg_rz( (icalendar_str!=0), "path");
	icalerror_check_arg_rz( (fset!=0), "fset");

	fset->icalendar_str = strdup(icalendar_str);
	fset->icalendar_str_work = fset->icalendar_str;
	fset->options = *options;

	icalerrorenum error;
	if( (error = icalstringset_read_cluster(fset)) != ICAL_NO_ERROR){
		icalstringset_free(set);
		return 0;
	}

    if (options->cluster) {
		fset->cluster = icalcomponent_new_clone(icalcluster_get_component(options->cluster));
		fset->changed = 1;
    }
    if (fset->cluster == 0) {
      fset->cluster = icalcomponent_new(ICAL_XROOT_COMPONENT);
    }
    return set;
}

char* icalstringset_read_from_icalendarstring(char *s, size_t size, void *d)
{
	icalstringset *set = d;

	strncpy(s, set->icalendar_str_work, size);
	char* p = s;

	const uint8_t CR = 0x0d;
	const uint8_t LF = 0x0a;

	for( p = s; p < s + size - 1; p++ ) {
		if( *p == CR ){
			p++;
			if( *p == LF ){
				p++;
			}
			break;
		}
		if( *p == LF ){
			p++;
			break;
		}
	}

    *p = '\0';

    if( *s == 0 ) {
		return 0;
    } else {
		set->icalendar_str_work += p - s;
		return s;
    }
}

icalerrorenum icalstringset_read_cluster(icalstringset* set)
{
    icalparser *parser;
  
    parser = icalparser_new();

    icalparser_set_gen_data(parser, set);
    set->cluster = icalparser_parse(parser, icalstringset_read_from_icalendarstring);
    icalparser_free(parser);

    if (set->cluster == 0 || icalerrno != ICAL_NO_ERROR){
		icalerror_set_errno(ICAL_PARSE_ERROR);
		/*return ICAL_PARSE_ERROR;*/
    }
  
    if (icalcomponent_isa(set->cluster) != ICAL_XROOT_COMPONENT){
		/* The parser got a single component, so it did not put it in
		   an XROOT. */
		icalcomponent *cl = set->cluster;
		set->cluster = icalcomponent_new(ICAL_XROOT_COMPONENT);
		icalcomponent_add_component(set->cluster,cl);
    }

    return ICAL_NO_ERROR;
}

void icalstringset_free(icalset* set)
{
    icalstringset *fset = (icalstringset*) set;
    icalerror_check_arg_rv((set!=0),"set");
    if (fset->cluster != 0){
		icalstringset_commit(set);
		icalcomponent_free(fset->cluster);
		fset->cluster=0;
    }

    if (fset->gauge != 0){
		icalgauge_free(fset->gauge);
		fset->gauge=0;
    }
	if(fset->icalendar_str != 0){
		free(fset->icalendar_str);
		fset->icalendar_str = 0;
    }
	if( fset->icalendar_str_work != 0) {
		// no need to free. Because fset->icalendar_str_work is just a pointer of fset->icalendar_str.
		fset->icalendar_str_work = 0;
	}
}

const char* icalstringset_icalendarstrring(icalset* set) {
    icalerror_check_arg_rz((set!=0),"set");

	return ((icalstringset*)set)->icalendar_str;
}

icalerrorenum icalstringset_commit(icalset* set)
{    
    return ICAL_NO_ERROR;
} 

void icalstringset_mark(icalset* set) {
    //icalerror_check_arg_rv((set!=0),"set");

    //((icalstringset*)set)->changed = 1;
}

/* manipulate the components in the set */

icalerrorenum icalstringset_add_component(icalset *set,
					icalcomponent* child)
{
    icalstringset *fset = (icalstringset*) set;

    icalerror_check_arg_re((set!=0),"set", ICAL_BADARG_ERROR);
    icalerror_check_arg_re((child!=0),"child",ICAL_BADARG_ERROR);

    icalcomponent_add_component(fset->cluster,child);

    icalstringset_mark(set);

    return ICAL_NO_ERROR;
}

icalerrorenum icalstringset_remove_component(icalset *set,
					   icalcomponent* child)
{
    icalstringset *fset = (icalstringset*) set;

    icalerror_check_arg_re((set!=0),"set",ICAL_BADARG_ERROR);
    icalerror_check_arg_re((child!=0),"child",ICAL_BADARG_ERROR);

    icalcomponent_remove_component(fset->cluster,child);

    icalstringset_mark(set);

    return ICAL_NO_ERROR;
}

int icalstringset_count_components(icalset *set,
				 icalcomponent_kind kind)
{
    icalstringset *fset = (icalstringset*) set;

    if (set == 0){
	icalerror_set_errno(ICAL_BADARG_ERROR);
	return -1;
    }

    return icalcomponent_count_components(fset->cluster,kind);
}

icalerrorenum icalstringset_select(icalset* set, icalgauge* gauge)
{
    icalstringset *fset = (icalstringset*) set;

    icalerror_check_arg_re(gauge!=0,"gauge",ICAL_BADARG_ERROR);

    fset->gauge = gauge;

    return ICAL_NO_ERROR;
}

void icalstringset_clear(icalset* set)
{
    icalstringset *fset = (icalstringset*) set;

    icalerror_check_arg_rv(set!=0,"set");

    fset->gauge = 0;
}

icalcomponent* icalstringset_fetch(icalset* set,const char* uid)
{
    icalstringset *fset = (icalstringset*) set;
    icalcompiter i;    

    icalerror_check_arg_rz(set!=0,"set");
    
    for(i = icalcomponent_begin_component(fset->cluster,ICAL_ANY_COMPONENT);
	icalcompiter_deref(&i)!= 0; icalcompiter_next(&i)){
	
		icalcomponent *this = icalcompiter_deref(&i);
		icalcomponent *inner;
		icalproperty *p;
		const char *this_uid;

		for(inner = icalcomponent_get_first_component(this,ICAL_ANY_COMPONENT);
			inner != 0;
			inner = icalcomponent_get_next_component(this,ICAL_ANY_COMPONENT)){

			p = icalcomponent_get_first_property(inner,ICAL_UID_PROPERTY);
			if ( p )
			{
				this_uid = icalproperty_get_uid(p);

				if(this_uid==0){
				//icalerror_warn("icalstringset_fetch found a component with no UID");
				continue;
				}

				if (strcmp(uid,this_uid)==0){
					return this;
				}
			}
		}
	}

    return 0;
}

int icalstringset_has_uid(icalset* set,const char* uid)
{
    assert(0); /* HACK, not implemented */
    return 0;
}

/******* support routines for icalstringset_fetch_match *********/

struct icalstringset_id{
    char* uid;
    char* recurrence_id;
    int sequence;
};

void icalstringset_id_free(struct icalstringset_id *id)
{
    if(id->recurrence_id != 0){
	free(id->recurrence_id);
    }

    if(id->uid != 0){
	free(id->uid);
    }
}


struct icalstringset_id icalstringset_get_id(icalcomponent* comp)
{
    icalcomponent *inner;
    struct icalstringset_id id;
    icalproperty *p;

    inner = icalcomponent_get_first_real_component(comp);
    
    p = icalcomponent_get_first_property(inner, ICAL_UID_PROPERTY);

    assert(p!= 0);

    id.uid = strdup(icalproperty_get_uid(p));

    p = icalcomponent_get_first_property(inner, ICAL_SEQUENCE_PROPERTY);

    if(p == 0) {
	id.sequence = 0;
    } else { 
	id.sequence = icalproperty_get_sequence(p);
    }

    p = icalcomponent_get_first_property(inner, ICAL_RECURRENCEID_PROPERTY);

    if (p == 0){
	id.recurrence_id = 0;
    } else {
	icalvalue *v;
	v = icalproperty_get_value(p);
	id.recurrence_id = icalvalue_as_ical_string_r(v);

	assert(id.recurrence_id != 0);
    }

    return id;
}


/* Find the component that is related to the given
   component. Currently, it just matches based on UID and
   RECURRENCE-ID */
icalcomponent* icalstringset_fetch_match(icalset* set, icalcomponent *comp)
{
    icalstringset *fset = (icalstringset*) set;
    icalcompiter i;    

    struct icalstringset_id comp_id, match_id;
    
    comp_id = icalstringset_get_id(comp);

    for(i = icalcomponent_begin_component(fset->cluster,ICAL_ANY_COMPONENT);
	icalcompiter_deref(&i)!= 0; icalcompiter_next(&i)){
	
	icalcomponent *match = icalcompiter_deref(&i);

	match_id = icalstringset_get_id(match);

	if(strcmp(comp_id.uid, match_id.uid) == 0 &&
	   ( comp_id.recurrence_id ==0 || 
	     strcmp(comp_id.recurrence_id, match_id.recurrence_id) ==0 )){

	    /* HACK. What to do with SEQUENCE? */

	    icalstringset_id_free(&match_id);
	    icalstringset_id_free(&comp_id);
	    return match;
	    
	}
	
	icalstringset_id_free(&match_id);
    }

    icalstringset_id_free(&comp_id);
    return 0;

}


icalerrorenum icalstringset_modify(icalset* set, icalcomponent *old,
				 icalcomponent *new)
{
	/* icalstringset *fset = (icalstringset*) set; */

    assert(0); /* HACK, not implemented */
    return ICAL_NO_ERROR;
}


/* Iterate through components */
icalcomponent* icalstringset_get_current_component (icalset* set)
{
    icalstringset *fset = (icalstringset*) set;

    icalerror_check_arg_rz((set!=0),"set");

    return icalcomponent_get_current_component(fset->cluster);
}


icalcomponent* icalstringset_get_first_component(icalset* set)
{
    icalcomponent *c=0;
    icalstringset *fset = (icalstringset*) set;

    icalerror_check_arg_rz((set!=0),"set");

    do {
		if (c == 0){
			c = icalcomponent_get_first_component(fset->cluster, ICAL_ANY_COMPONENT);
		} else {
			c = icalcomponent_get_next_component(fset->cluster, ICAL_ANY_COMPONENT);
		}

		if(c != 0 && (fset->gauge == 0 ||
				  icalgauge_compare(fset->gauge, c) == 1)){
			return c;
		}
    } while(c != 0);
    return 0;
}

icalcomponent* icalstringset_get_next_component(icalset* set)
{
    icalstringset *fset = (icalstringset*) set;
    icalcomponent *c;

    icalerror_check_arg_rz((set!=0),"set");
    
    do {
		c = icalcomponent_get_next_component(fset->cluster, ICAL_ANY_COMPONENT);

		if(c != 0 && (fset->gauge == 0 || icalgauge_compare(fset->gauge,c) == 1)){
		    return c;
		}
	
    } while(c != 0);
    
    
    return 0;
}
/*
icalsetiter icalstringset_begin_component(icalset* set, icalcomponent_kind kind, icalgauge* gauge)
{
    icalsetiter itr = icalsetiter_null;
    icalcomponent* comp = NULL;
    icalcompiter citr;
    icalstringset *fset = (icalstringset*) set;

    icalerror_check_arg_re((set!=0), "set", icalsetiter_null);

    itr.gauge = gauge;

    citr = icalcomponent_begin_component(fset->cluster, kind);
    comp = icalcompiter_deref(&citr);

    while (comp != 0) {
	comp = icalcompiter_deref(&citr);
	if (gauge == 0 || icalgauge_compare(itr.gauge, comp) == 1) {
	    itr.iter = citr;
	    return itr;
	}
	comp =  icalcompiter_next(&citr);
    }

    return icalsetiter_null;
}
*/

icalsetiter icalstringset_begin_component(icalset* set, icalcomponent_kind kind, icalgauge* gauge)
{
    icalsetiter itr = icalsetiter_null;
    icalcomponent* comp = NULL;
    icalcompiter citr;
    icalstringset *fset = (icalstringset*) set;
    struct icaltimetype start, next;
    icalproperty *dtstart, *rrule, *prop, *due;
    struct icalrecurrencetype recur;
    int g = 0;

    icalerror_check_arg_re((set!=0), "set", icalsetiter_null);

    itr.gauge = gauge;

    citr = icalcomponent_begin_component(fset->cluster, kind);
    comp = icalcompiter_deref(&citr);

    if (gauge == 0) {
        itr.iter = citr;
        return itr;
    }

    while (comp != 0) {

        /* check if it is a recurring component and with guage expand, if so
           we need to add recurrence-id property to the given component */
        rrule = icalcomponent_get_first_property(comp, ICAL_RRULE_PROPERTY);
        g = icalgauge_get_expand(gauge);

        if (rrule != 0
            && g == 1) {

            recur = icalproperty_get_rrule(rrule);
            if (icalcomponent_isa(comp) == ICAL_VEVENT_COMPONENT) {
                dtstart = icalcomponent_get_first_property(comp, ICAL_DTSTART_PROPERTY);
                if (dtstart)
                    start = icalproperty_get_dtstart(dtstart);
            } else if (icalcomponent_isa(comp) == ICAL_VTODO_COMPONENT) {
                    due = icalcomponent_get_first_property(comp, ICAL_DUE_PROPERTY);
                    if (due)
                        start = icalproperty_get_due(due);
            }

            if (itr.last_component == NULL) {
                itr.ritr = icalrecur_iterator_new(recur, start);
                next = icalrecur_iterator_next(itr.ritr);
                itr.last_component = comp;
            }
            else {
                next = icalrecur_iterator_next(itr.ritr);
                if (icaltime_is_null_time(next)){
                    itr.last_component = NULL;
                    icalrecur_iterator_free(itr.ritr);
                    itr.ritr = NULL;
                    return icalsetiter_null;
                } else {
                    itr.last_component = comp;
                }
            }

            /* add recurrence-id to the component
               if there is a recurrence-id already, remove it, then add the new one */
            prop = icalcomponent_get_first_property(comp, ICAL_RECURRENCEID_PROPERTY);
	    if (prop)
                icalcomponent_remove_property(comp, prop);
            icalcomponent_add_property(comp, icalproperty_new_recurrenceid(next));

        }

        if (gauge == 0 || icalgauge_compare(itr.gauge, comp) == 1) {
            /* matches and returns */
            itr.iter = citr;
            return itr;
        }

        /* if there is no previous component pending, then get the next component */
        if (itr.last_component == NULL)
            comp =  icalcompiter_next(&citr);
    }

    return icalsetiter_null;
}

icalcomponent* icalstringsetiter_to_next(icalset* set, icalsetiter* i)
{

    icalcomponent* c = NULL;
    /* icalstringset *fset = (icalstringset*) set; */
    struct icaltimetype start, next;
    icalproperty *dtstart, *rrule, *prop, *due;
    struct icalrecurrencetype recur;
    int g = 0;


    do {
        c = icalcompiter_next(&(i->iter));

        if (c == 0) continue;
        if (i->gauge == 0) return c;


        rrule = icalcomponent_get_first_property(c, ICAL_RRULE_PROPERTY);
        g = icalgauge_get_expand(i->gauge);

        /* a recurring component with expand query */
        if (rrule != 0
            && g == 1) {

            recur = icalproperty_get_rrule(rrule);

            if (icalcomponent_isa(c) == ICAL_VEVENT_COMPONENT) {
                dtstart = icalcomponent_get_first_property(c, ICAL_DTSTART_PROPERTY);
                if (dtstart)
                    start = icalproperty_get_dtstart(dtstart);
            } else if (icalcomponent_isa(c) == ICAL_VTODO_COMPONENT) {
                due = icalcomponent_get_first_property(c, ICAL_DUE_PROPERTY);
                if (due)
                    start = icalproperty_get_due(due);
            }

            if (i->ritr == NULL) {
                i->ritr = icalrecur_iterator_new(recur, start);
                next = icalrecur_iterator_next(i->ritr);
                i->last_component = c;
            } else {
                next = icalrecur_iterator_next(i->ritr);
                if (icaltime_is_null_time(next)) {
                    /* no more recurrence, returns */
                    i->last_component = NULL;
                    icalrecur_iterator_free(i->ritr);
                    i->ritr = NULL;
                    return NULL;
                } else {
                    i->last_component = c;
                }
            }
        }

        /* add recurrence-id to the component
         * if there is a recurrence-id already, remove it, then add the new one */
	prop = icalcomponent_get_first_property(c, ICAL_RECURRENCEID_PROPERTY);
        if (prop)
            icalcomponent_remove_property(c, prop);
        icalcomponent_add_property(c, icalproperty_new_recurrenceid(next));

        if(c != 0 && (i->gauge == 0 ||
                        icalgauge_compare(i->gauge, c) == 1)){
            return c;
        }
    } while (c != 0);

    return 0;

}
